@@ -16,8 +16,8 @@
 #' @export mkBuildmer
 mkBuildmer <- setClass('buildmer',slots=list(model='ANY',p='list',anova='ANY',summary='ANY'))
 
-#' @S3method show buildmer
 #' @import methods
+#' @S3method show buildmer
 show.buildmer <- function (object) {
 	methods::show(object@model)
 	if (length(object@p$messages)) {
@@ -28,7 +28,6 @@ show.buildmer <- function (object) {
 setMethod('show','buildmer',show.buildmer)
 
 #' @S3method anova buildmer
-#' @import stats
 anova.buildmer <- function (object,...) {
 	if (length(object@p$messages)) warning(object@p$messages)
 	dots <- list(...)
@@ -61,11 +60,11 @@ anova.buildmer <- function (object,...) {
 		ddf <- 'Wald'
 	}
 
+	if (is.null(type)) type <- 3
 	if (ddf %in% c('Wald','lme4')) {
 		table <- if (inherits(object@model,'lmerModLmerTest')) stats::anova(object@model,ddf='lme4',type=type) else stats::anova(object@model)
 		if (ddf == 'Wald') {
-			table <- calcWald(table,4,sqrt=T)
-			if (is.null(type)) type <- 3
+			table <- calcWald(table,4,col.df=1)
 			attr(table,'heading') <- paste('ANOVA based on type',utils::as.roman(type),'SS\n(p-values based on the Wald chi-square approximation)')
 		}
 	} else {
